/* evtgen lexer.
   Copyright (C) 2003 Julien Olivain and LSV, CNRS UMR 8643 & ENS Cachan.

   This file is part of evtgen.

   evtgen is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   evtgen is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with evtgen; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

%{

/*
** parser.l -- parser for event_generator
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "evtgen.h"

#include "parser.tab.h"

/*
** since the maxline limit is usually 4096 and quoted string can't
** be defined in multiple lines, we can't have a qstr bigger than 4096.
*/
#define QSTRMAX 4096

#define QSTR_PUTC(c) \
  do { \
    if (qstr_sz_g >= (QSTRMAX - 1)) { \
      printf("%24s:%-4d: quoted string too long\n", __FILE__, __LINE__); \
      exit(EXIT_FAILURE); \
     } \
     qstr_sz_g++; \
     *qstr_ptr_g++ = (c); \
  } while (0)

unsigned int parserlineno_g = 1;

/* static values for escape character interpretation and
** quoted string construction. */
static char   qstr_buf_g[QSTRMAX];
static char  *qstr_ptr_g;
static size_t qstr_sz_g;

/* Add this line with BYacc */
/* extern int parserlval; **/

%}

/* eXclusive state COMMENT */
%x COMMENT QSTR

%option nounput

%%

[\t ]+ /* ignore whitespace */ ;
"\n" { ++parserlineno_g; }

 /* reserved keywords and operators */

 /* handle comments */
"/*"       { BEGIN COMMENT; }
<COMMENT>. ;
<COMMENT>\n   { ++parserlineno_g; }
<COMMENT>"*/" { BEGIN INITIAL; }
<COMMENT>"/*" { DPRINTF( ("warning! '/*' in comment.\n") ); }

 /* built-in variable name */
$[A-Za-z_][A-Za-z0-9_]* {
  parserlval.str = strdup(parsertext + 1);
  /* DPRINTF( ("BUILTIN_VAR %s\n", parsertext) ); */
  return (BUILTIN_VAR);
}

 /* varlist name */
@[a-zA-Z_]+[a-zA-Z0-9_]* {
  parserlval.str = strdup(parsertext + 1);
  /* DPRINTF( ("VAR %s\n", parsertext) ); */
  return(VAR);
}

 /* symbol names (same as cdecls) */
[A-Za-z_][A-Za-z0-9_]* {
  parserlval.sym.line = parserlineno_g;
  parserlval.sym.name = strdup(parsertext);
  /* DPRINTF( ("SYMNAME %s\n", parsertext) ); */
  return (SYM);
}

 /* float numbers */
 /* complete float "([0-9]+|([0-9]*\.[0-9]+)([eE][-+]?[0-9]+)?)" */
 /* simpler [0-9]+(\.[0-9]+)? */
 /* prob (0\.[0-9]+|1\.0) */
([0-9]+|([0-9]*\.[0-9]+)([eE][-+]?[0-9]+)?) {
  parserlval.prob = atof(parsertext);
  /* DPRINTF( ("PROB %5.3f\n", parserlval.prob) ); */
  return (PROB);
}

 /* quoted string handling */
 /* a double quote is the begin of a quoted string */
\" {
  qstr_sz_g = 0;
  qstr_ptr_g = qstr_buf_g;
  BEGIN QSTR;
}

  /* end of the string */
<QSTR>\" {
  BEGIN INITIAL;
  *qstr_ptr_g = '\0';
  parserlval.str = strdup(qstr_buf_g);
  return (STR);
}

  /* newline in qstr must be specified as \n */
<QSTR>\n {
  DPRINTF( ("ERROR: newline in multiline quoted strings must be escaped.\n") );
  return (0);
}

  /* octal sequence handling */
<QSTR>\\[0-7]{1,3} {
  unsigned int res;

  sscanf(yytext + 1, "%o", &res);
  if (res > 0xFF)
    {
      DPRINTF( ("ERROR: bad octal value (%i)\n", res) );
      return (0);
    }
  QSTR_PUTC(res);
}

  /* other illegal values */
<QSTR>[0-9]+ {
  DPRINTF( ("ERROR: bad octal value (%s)\n", yytext) );
  return (0);
}

  /* new line */
<QSTR>\\n { QSTR_PUTC('\n'); }

  /* tabulation */
<QSTR>\\t { QSTR_PUTC('\t'); }

  /* carriage return */
<QSTR>\\r { QSTR_PUTC('\r'); }

  /* backspace */
<QSTR>\\b { QSTR_PUTC('\b'); }

  /* form feed */
<QSTR>\\f { QSTR_PUTC('\f'); }

  /* other escaped characters (\", \\n, ...) */
<QSTR>\\(.|\n) { QSTR_PUTC(yytext[1]); }

  /* traditional string */
<QSTR>[^\\\n\"]+ {
  char *yptr;

  yptr = yytext;
  while (*yptr)
    QSTR_PUTC(*yptr++);
}

. {
  return (parsertext[0]);
}

%%

/*
** yywrap()
*/

int
parserwrap(void)
{
  return (1);
}
