/* evtgen string hash functions.
   Copyright (C) 2003 Julien Olivain and LSV, CNRS UMR 8643 & ENS Cachan.

   This file is part of evtgen.

   evtgen is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   evtgen is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with evtgen; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
** strhash.c -- string hash utility functions
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "evtgen.h"

#include "strhash.h"

strhash_t *
new_strhash(size_t hsize)
{
  strhash_t *h;

  h = xmalloc(sizeof (strhash_t));
  h->size = hsize;
  h->elmts = 0;
  h->hash = DEFAULT_STRHASH_FUNCTION;
  h->htable = xzmalloc(hsize * sizeof (strhash_elmt_t *));

  return (h);
}

void
clear_strhash(strhash_t *hash, void (*elmt_free)(void *e))
{
  int i;
  strhash_elmt_t *tmp;
  strhash_elmt_t *tmp_next;

  for (i = 0; i < hash->size; i++)
    {
      tmp = hash->htable[i];
      while (tmp)
        {
          tmp_next = tmp->next;
          if (elmt_free)
            elmt_free(tmp->data);
          free(tmp);
          tmp = tmp_next;
        }
    }
  hash->elmts = 0;
  memset(hash->htable, 0, hash->size * sizeof (void *));
}


void
free_strhash(strhash_t *hash, void (*elmt_free)(void *e))
{
  int i;
  strhash_elmt_t *tmp;
  strhash_elmt_t *tmp_next;

  for (i = 0; i < hash->size; i++)
    {
      tmp = hash->htable[i];
      while (tmp)
        {
          tmp_next = tmp->next;
          if (elmt_free)
            elmt_free(tmp->data);
          free(tmp);
          tmp = tmp_next;
        }
    }

  free(hash->htable);
  free(hash);
}

#if 0
void
strhash_resize(strhash_t *hash, size_t newsize)
{
  strhash_elmt_t **old_htable;
  int i;

  /* XXX: Not tested !!! */
  printf("Not tested !...\n");

  old_htable = hash->htable;
  hash->htable = xzmalloc(newsize * sizeof (strhash_elmt_t *));

  for (i = 0; i < hash->size; ++i)
    {
      strhash_elmt_t *tmp;

      for (tmp = old_htable[i]; tmp; tmp = tmp->next)
        {
          unsigned long hcode;

          hcode = hash->hash(tmp->key) % newsize;
          tmp->next = hash->htable[hcode];
          hash->htable[hcode] = tmp;
        }
    }

  free(old_htable);
  hash->size = newsize;
}
#endif /*0*/

void
strhash_add(strhash_t *hash, void *data, char *key)
{
  strhash_elmt_t *elmt;
  unsigned long hcode;

  elmt = xmalloc(sizeof (strhash_elmt_t));
  elmt->key = key;
  elmt->data = data;

  hcode = hash->hash(key) % hash->size;
  elmt->next = hash->htable[hcode];
  hash->htable[hcode] = elmt;
  hash->elmts++;
}

void *
strhash_get(strhash_t *hash, char *key)
{
  strhash_elmt_t *elmt;

  elmt = hash->htable[hash->hash(key) % hash->size];
  for (; elmt; elmt = elmt->next)
    {
      if (!strcmp(key, elmt->key))
        return (elmt->data);
    }

  return (NULL);
}

int
strhash_walk(strhash_t *hash, int (func)(void *elmt, void *data), void *data)
{
  int i;

  for (i = 0; i < hash->size; i++)
    {
      strhash_elmt_t *tmp;
      int status;

      for (tmp = hash->htable[i]; tmp; tmp = tmp->next)
        {
          if ((status = func(tmp->data, data)) != 0)
            return (status);
        }
    }
  return (0);
}


/*
** fast hash function samples
*/

unsigned long
strhash_pjw(char *key)
{
  unsigned long h;
  unsigned long g;

  h = 0;
  while (*key)
    {
      h = (h << 4) + *key++;
      if ((g = h & 0xF0000000U) != 0)
        {
          h = h ^ (g >> 24);
          h = h ^ g;
        }
    }

  return (h);
}

unsigned long
strhash_pow(char *key)
{
  unsigned long hcode;

  hcode = 0;
  while (*key)
    {
      hcode += (*key) * (*key);
      key++;
    }

  return (hcode);
}

unsigned long
strhash_x65599(char *key)
{
  unsigned long hcode;

  hcode = 0;
  while (*key)
    hcode += *key++ * 65599;

  return (hcode);
}

unsigned long
strhash_quad(char *key)
{
  int quads;
  int i;
  unsigned long h;
  unsigned long *val;

  val = (unsigned long *) key;
  quads = strlen(key) / sizeof (unsigned long);
  h = 0;
  for (i = 0; i < quads; i++)
    h += *val++;

  return (h);
}

int
strhash_collide_count(strhash_t *hash)
{
  int count;
  int i;

  count = 0;
  for (i = 0; i < hash->size; i++)
    {
      strhash_elmt_t *tmp;

      for (tmp = hash->htable[i]; tmp; tmp = tmp->next)
        if (tmp->next)
          count++;
    }

  return (count);
}

#if 0
int
print_elmt(void *elmt, void *dummy)
{
  printf("%i\n", (int)elmt);

  return (0);
}

void
hash_test(void)
{
  strhash_t *h;
  char buf[1024];
  int i;

  h = new_hash(2053);
  h->hash = hash_pow;
  for (i = 0; i < 1000; i++)
    {
      sprintf(buf, "%i", i * 65599);
      hash_add(h, (void *)(i * 65599), buf, strlen(buf));
    }
  hash_walk(h, print_elmt, NULL);

  printf("collides : %i\n", hash_collide_count(h));
}

void
fprintf_hash_info(FILE *fp, strhash_t *h)
{
  fprintf(fp, "htable = %p\n", h->htable);
  fprintf(fp, "size = %i\n", h->size);
  fprintf(fp, "elmts = %i\n", h->elmts);
  fprintf(fp, "hfunc = %p\n", h->hash);
}

#endif /* DEBUG */
